@@ -61,20 +61,17 @@ def if_equal_to_the_nearest_int_convert_to_int_else_float(num):
 
 
 def factors_of_x(x, y=1):
-    type_x = type(x)
-    type_y = type(y)
+
     x = convert_to_int_or_float(x)
     y = convert_to_int_or_float(y)
-    x_is_integer = isinstance(x, int)
-    y_is_integer = isinstance(y, int)
 
-    if not x_is_integer:
-        raise Exception('x must be an integer. Now, type_x = {}'.format(type_x))
+    if isinstance(x, float):
+        raise TypeError('Type of x must be int')
 
-    if not y_is_integer:
-        raise Exception('y must be an integer. Now, type_y = {}'.format(type_y))
+    if isinstance(y, float):
+        raise TypeError('Type of y must be int')
 
-    factors = np.empty(0, int)
+    factors = np.empty(0, dtype='i')
 
     for i in range(y, x + 1):
         if x % i == 0:
